class Constants(object):

    PARENT = "origin_record"
    CIRCULAR = "circular"
    LINEAR = "linear"
    TRANSFORMATION = "modification"
    OLD_KEY = "original_record_id"
    PSEUDOCIRCULAR = "pseudocircularized"
    COPY_RECORD = "copy_record"
    TOPOLOGY = "topology"

    class NULL(object):
        """A not NONE class."""
